// summarise.c -- dialserver log parser and summariser
//
// This reads the log sequentially, and reports summary information 
// about the usage of the dialserver web interface.  The log is 
// assumed to be in as-happened order.
//
// Copyright (c) 2001,2002 Andrew McGill and Leading Edge Business Solutions
//
// This file is part of dialserver.
//
// dialserver is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// dialserver is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.
//
// You should have received a copy of the GNU General Public License along with
// dialserver; if not, write to the Free Software Foundation, Inc., 59 Temple
// Place, Suite 330, Boston, MA  02111-1307  USA

#include "common.h"
#include "util.h"
#include "conf.h"

// Dialserver usage report 27 Jul - 28 Jul
//
// Call summary --
//
// Date    Calls   Total time  Estimated cost
// 27 Jul    2        6.22     R3.12
// 28 Jul    3        7.21     R3.62
// Total     5       13.23     R6.74
//
// User summary --
// User : joe
// Date    Logins  Total time  Sites  Random site
// 27 Jul    2        6.22     12     http://www.jaberwock.com/henry
// 28 Jul    3        7.21     231    http://www.blunderbus.net/jab/blank.gif
// Total     5       13.23     243
//
// User : mail
// Date    Logins  Total time   Random site
// 27 Jul    2        6.22     Login
// 28 Jul    3        7.21     Login
// Total     5       13.23     

typedef struct _CALLSUMMARY {
	char date[10];		// The date for which this is a summary
	int calls;		// The number of calls summarised
	int time;		// Total time
	int cost;		// Total cost (estimated)
} CALLSUMMARY;

typedef struct _MONTHLIST {
	char month[10];		// The date for which this is a summary
} MONTHLIST;

// For each day, there's one of these for each user
typedef struct _USERSUMMARY {
	char key[32];		// User and date (key for searching)
	char ident[22];		// Who
	char date[10];		// The date for which this is a summary
	int calls;		// How many times he logged in
	int starttime, endtime;		// minute of the day start and finish
	int time;		// Total time
	int printed;		// Set to 1 when this record has been printed
} USERSUMMARY;

struct _LINKEDLIST *calls;	// Data is CALLSUMMARY
struct _LINKEDLIST *monthlist;	// Data is MONTHLIST
struct _LINKEDLIST *users;	// Data is USERSUMMARY

// LINKEDLIST -- a linked list.  There are a whole bunch of list_* functions
// which do the real job ...
typedef struct _LINKEDLIST {
	struct _LINKEDLIST *next;
	void *data;
} LINKEDLIST; 

// Find a list item which compares to the supplied item using a specified
// comparison function
LINKEDLIST *list_find(LINKEDLIST *base, void *data, int(*comparefn)(void *data1, void *data2))
{
	while (base) {
		if (base->data == data)
			return base;
		if (comparefn(base->data,data)==0) 
			return base;
		base = base->next;
	}
	return NULL;
}

// Count the number of items in the list
int list_count(LINKEDLIST *base)
{
	int rc;
	rc = 0;
	while (base) {
		rc++;
		base=base->next;
	}
	return rc;
}

// Free string list and strings
void list_free(LINKEDLIST *item)
{
	LINKEDLIST *next;
	while (item) {
		next=item->next;
		free(item->data);
		free(item);
		item=next;
	}
}

// Adds the supplied pointer to the list
LINKEDLIST *list_add(LINKEDLIST **base, void *data)
{
	LINKEDLIST *add;
	LINKEDLIST **pointer;

	// Add to the end of the list
	pointer = base;
	while (*pointer)
		pointer=&((*pointer)->next);
	// Allocate some memory
	add=xmalloc(sizeof(LINKEDLIST));
	add->data = data;
	add->next = NULL;
	*pointer = add;
	return add;
}


// OTHER STUFF

int mystrcmp(void *c1, void *c2)
{
	return strcmp(c1,c2);
}

// Convert a date to a month and list it ...
void add_month(char *date)
{
	char month[10];
	size_t span;
	MONTHLIST *themonth;
	LINKEDLIST *item;

	strncpy(month,date,sizeof(month));
	span=strcspn(month," ");
	if (span==0 || span>=sizeof(month)) return;
	month[span]=0;

	if (!(item = list_find(monthlist,month,mystrcmp))) {
		themonth = xcalloc(1,sizeof(MONTHLIST));
		item = list_add(&monthlist,themonth);
		strncpy(themonth->month,month,sizeof(themonth->month)-1);
	}
}

// Add a call to the list
void add_call(char *date,  int time)
{
	LINKEDLIST *item;
	CALLSUMMARY *call;
	// Find or create a record for the date in calls
	if ((item = list_find(calls,date,mystrcmp))) {
		call = item->data;	
	}
	else {
		call = xcalloc(1,sizeof(CALLSUMMARY));
		item = list_add(&calls,call);
		strncpy(call->date,date,sizeof(call->date)-1);
	}
	if (time<180) {
		call->cost += 50;
	}
	else {
		call->cost += time*50/180;
	}
	call->time += time;
	call->calls++;
}

// Add a call to the user's list
void add_usercall(char *date, char *ident, int time, int endtime)
{
	char key[23];
	int starttime;
	
	LINKEDLIST *item;
	USERSUMMARY *u;

	if (time > 10800 || time < 0 ) { /* sanity check - workaround buggy accounting elsewhere */
		time = 0;
	}
	starttime = endtime - time/60;
	if (starttime<=0) starttime=1;

	snprintf(key,sizeof(key),"%s %s",ident,date);

	// Find or create a record for the date in calls
	if ((item = list_find(users,key,mystrcmp))) {
		u = item->data;	
	}
	else {
		u = xcalloc(1,sizeof(USERSUMMARY));
		item = list_add(&users,u);
		strncpy(u->key, key,  sizeof(u->key)-1);
		strncpy(u->ident,ident,sizeof(u->ident)-1);
		strncpy(u->date,date, sizeof(u->date)-1);
	}
	u->calls++;
	u->time += time;
	if (! u-> starttime ) u->starttime = starttime;
	u->endtime = endtime; 
}

// Process a line from the log file
void sum_parseline(char *line)
{
	// Skip month, day and time for now
	char *month, *day, *hour, *min, *sec, *word;
	char *x, *y, *z;
	int time;

	if (!(month = strtok(line," "))) return;
	if (!(day   = strtok(NULL," "))) return;
	day[-1]=' ';  // Put back the space
	if (!(hour  = strtok(NULL,":"))) return;
	if (!(min   = strtok(NULL,":"))) return;
	if (!(sec   = strtok(NULL," "))) return;
	if (!(word  = strtok(NULL," "))) return;

	if (strcasecmp(word,"call-duration")==0) {
		if (!(x = strtok(NULL,":. "))) return;   // Minutes	
		if (!(y = strtok(NULL," "))) return;	// Seconds
		time = atoi(x)*60+atoi(y);
		add_month(month);
		add_call(month,time);
	}
	else if (strcasecmp(word,"logout")==0) {
		if (!(x = strtok(NULL," "))) return;    // User    lyn
		if (!(y = strtok(NULL," "))) return;	// Reason  dropped
		if (!(y = strtok(NULL,":. "))) return;	// Minutes 05:58
		if (!(z = strtok(NULL," "))) return;	// Seconds
		time = atoi(y)*60+atoi(z);
		add_month(month);
		add_usercall(month,x,time, atoi(min)+atoi(hour)*60);
	}
	// TODO: make a list of IP addresses that each user has authenticated
	// from in a day
	//else if (strcasecmp(word,"start")==0) { /* start 127.0.0.1 (root) */
	//	if (!(x = strtok(NULL," "))) return;    // ip      127.0.0.1
	//	if (!(y = strtok(NULL,"("))) return;	// (
	//	if (!(y = strtok(NULL,")"))) return;	// user    root
	//	add_userip(month,y,x);
	//}
}

// Parseline
int readlogfile(char *filename)
{
	FILE *f;
	char line[200];

	f=fopen(filename,"r");
	if (!f) {
		perror("open");
		return -1;
	}
	while (!feof(f)) {
		if (fgets(line,sizeof(line),f) == NULL) {
			break;
		}
		if (ferror(f)) goto oops;
		sum_parseline(line);
	}
	if (ferror(f)) goto oops;
	if (fclose(f)) {
		perror("close");
		return -1;
	}
	return 0;

oops:
	perror("fgets");
	fclose(f);
	return -1;
}

// Ooh! - this is a comment that just floats around


// Time to printable string [hh:]mm:ss
char *xtimetostring(char *buf, int buflen, int seconds, int spaces, char *sep)
{
	if (seconds >= 3600) {
		snprintf(buf,buflen,"%3d.%.2d%s%.2d",
			seconds / 3600,
			(seconds % 3600) / 60 ,
			sep,
			(seconds % 60));
	}
	else {
		snprintf(buf,buflen,( spaces ? "    %2d%s%.2d" : "%d%s%.2d"),
			(seconds % 3600) / 60 ,
			sep,
			(seconds % 60));
	}
	return buf;
}
char *timetostring(char *buf, int buflen, int seconds)
{
	return xtimetostring(buf,buflen,seconds,1,":");
}
char *timetostr(char *buf, int buflen, int seconds)
{
	return xtimetostring(buf,buflen,seconds,0,"h");
}

// Print a user record and exit
int printuserrecords(char *month)
{
	LINKEDLIST *list;
	USERSUMMARY *user, *theuser, usertotal;
	char timebuf[32],timebuf2[32],timebuf3[32];

	// Print the totals for all calls
	memset(&usertotal,0,sizeof(usertotal));
	list=calls;
	theuser=NULL;
	for (list=users; list; list=list->next) {
		user = list->data;
		if ( user->printed ||
			strncmp(month,user->date,strlen(month))!=0 ) continue;
		if (!theuser) {
			theuser = user;
			fprintf(stdout,"%s user summary: %s --\n",month,user->ident);
			fprintf(stdout,"Date\tLogins\t  Time\n");
		}
		if (theuser && strcasecmp(user->ident,theuser->ident)==0 ) {
			user->printed = 1;
			fprintf(stdout,"%s\t%3d\t%s\t%s (%s-%s)\n",
				user->date, user->calls,
				timetostring(timebuf,sizeof(timebuf),user->time),
				user->ident,
				timetostr(timebuf2,sizeof(timebuf2),user->starttime),
				timetostr(timebuf3,sizeof(timebuf3),user->endtime)
				);
			usertotal.calls+= user->calls;
			usertotal.time += user->time;
		}
	}
	if (!theuser) return 0;
	fprintf(stdout,"%s   \t%3d\t%s\t%s (total)\n\n",
		month,
		usertotal.calls,
		timetostring(timebuf,sizeof(timebuf),usertotal.time),
		theuser->ident);
	return 1;   // Return 1 because we did do something 
}

// Show the results
void printcallsummary(char *month)
{
	LINKEDLIST *list;
	CALLSUMMARY *call, calltotal;
	char timebuf[32];

	// Print the totals for all calls
	memset(&calltotal,0,sizeof(calltotal));
	fprintf(stdout,"%s call summary\n",month);
	fprintf(stdout,"Date\tCalls\t  Time\tEst.cost\n");
	list=calls;
	for (list=calls; list; list=list->next) {
		call = list->data;
		if ( strncmp(month,call->date,strlen(month))!=0 ) continue;
		fprintf(stdout,"%s\t%3d\t%s\t%3d.%.2d\n",
			call->date, call->calls,
			timetostring(timebuf,sizeof(timebuf),call->time),
			call->cost/100, call->cost%100);
		calltotal.time += call->time;
		calltotal.cost += call->cost;
		calltotal.calls+= call->calls;
	}
	fprintf(stdout,"%s   \t%3d\t%s\t%3d.%.2d\n\n",
		month,
		calltotal.calls,
		timetostring(timebuf,sizeof(timebuf),calltotal.time),
		calltotal.cost/100, calltotal.cost%100);
}

// Blah
void printmonthresults(void)
{
	LINKEDLIST *list;
	MONTHLIST *month;

	fprintf(stdout,"CALL SUMMARIES FOR EACH MONTH\n\n");
	fprintf(stdout,"Costs are estimated cost at 50c per 3 minutes\n"
		"(Make an adjustment for actual call rates)\n"
		"\n");
	for (list=monthlist; list; list=list->next) {
		month = list->data;
		printcallsummary(month->month);
	}

	fprintf(stdout,"\n\nUSER SUMMARIES FOR EACH MONTH\n\n");
	for (list=monthlist; list; list=list->next) {
		month = list->data;
		while (printuserrecords(month->month)) {
			// do nothing in particular, until all is printed
		};
	}
}

int main(int argc, char *argv[])
{
	int rc=0,a;
	if (argc<2) {
		dlr_loadconfig(CONFIGFILE,"dialserver");
		rc=readlogfile(conf.logfile);
	} 
	else {
		for (a=1; a<argc; a++) {
			rc=readlogfile(argv[a]);
		}
	}
	if (rc != 0) {
		fprintf(stderr,"Usage: %s [logfile]\n"
			"(Default log file is as specified in %s)\n",
			argv[0],CONFIGFILE);
		return 1;
	}
	printmonthresults();
	return 0;
}

